# Nymph Client - collaborative app data

[![Latest Stable Version](https://img.shields.io/npm/v/nymph-client.svg)](https://www.npmjs.com/package/nymph-client) [![Open Issues](https://img.shields.io/github/issues/sciactive/nymph-client.svg)](https://github.com/sciactive/nymph-client/issues) [![License](https://img.shields.io/github/license/sciactive/nymph-client.svg)]()

Nymph is an object data store that is easy to use in JavaScript and PHP.

## Installation

### Automatic Setup

The fastest way to start building a Nymph app is with the Nymph App Template.

[Nymph App Template](https://github.com/hperrin/nymph-template)

### Manual Installation

You can install Nymph Client with NPM.

```sh
npm install --save nymph-client
```

This repository is the JavaScript client for browsers. You can find UMD format modules in the /lib directory. There is also a **[Node.js client](https://github.com/sciactive/nymph-client-node)**. For more information, you can see the [main Nymph repository](https://github.com/sciactive/nymph).

## Setting up Nymph in a Browser

<div dir="rtl">Quick Setup with NPM</div>

```sh
npm install --save nymph-client
```
```html
<head>
  <!-- ... -->
  NymphOptions = {
    restURL: 'https://yournymphrestserver/path/to/your/rest.php',
    pubsubURL: 'wss://yournymphpubsubserver:8080',
    rateLimit: 100
  };
  <script src="node_modules/nymph-client/lib/Nymph.js"></script>
  <script src="node_modules/nymph-client/lib/Entity.js"></script>
  <script src="node_modules/nymph-client/lib/PubSub.js"></script>
  <script src="path/to/your/entity/js/Todo.js"></script>
</head>
```
```js
// Now, in your module that requires Nymph...
// (something like `import Nymph from 'Nymph';`)

const myTodo = new Todo();
myTodo.set({
  name: "This is a new todo!",
  done: false
});
myTodo.save().then(() => {
  alert("You've got new todos!");
});

// ...

let subscription = Nymph.getEntities({"class": Todo.class}, {"type": '&', "!tag": 'archived'}).subscribe((newTodos) => {
  if (newTodos !== undefined) {
    var todos = this.get('todos');
    Nymph.updateArray(todos, newTodos);
    Nymph.sort(todos, this.get('uiSort'));
    this.set({todos: todos});
  }
}, null, (count) => {
  this.set({userCount: count});
});
```

For a thorough step by step guide to setting up Nymph on your own server, visit the [Setup Guide](https://github.com/sciactive/nymph/wiki/Setup-Guide).

## API Docs

Check out the [API Docs in the wiki](https://github.com/sciactive/nymph/wiki/API-Docs).
